-- Initialization

local GearEvaluator = Rock:NewAddon("GearEvaluator", "LibRockDB-1.0", "LibRockConfig-1.0", "LibRockConsole-1.0", "LibRockEvent-1.0", "LibRockModuleCore-1.0", "LibFuBarPlugin-3.0")

GearEvaluator:SetDatabase("GearEvaluatorDB", "GearEvaluatorPCDB")

local slotnames = {"HeadSlot",
  "NeckSlot",
  "ShoulderSlot",
  "BackSlot",
  "ChestSlot",
  "ShirtSlot",
  "TabardSlot",
  "WristSlot",
  "HandsSlot",
  "WaistSlot",
  "LegsSlot",
  "FeetSlot",
  "Finger0Slot",
  "Finger1Slot",
  "Trinket0Slot",
  "Trinket1Slot",
  "MainHandSlot",
  "SecondaryHandSlot",
  "RangedSlot",
  "AmmoSlot",
  --"Bag0Slot",
  --"Bag1Slot",
  --"Bag2Slot",
  --"Bag3Slot"
}

-- Functions

function GearEvaluator:OnInitialize()
  if (not self.scores) then
    self.scores = {}
  end

  if (not self.errors) then
    self.errors = {}
  end

  if (not self.evaluators) then
    self.evaluators = {}
  end

  if (not self.attrFunctions) then
    self.attrFunctions = {}
  end
  
  self:SetDatabaseDefaults('profile', { evaluators = {} })

  self:CreateFrames()
  
  self.evaluatorConfigTable = {
    add = {
      type = 'string',
      name = "Add evaluator",
      desc = "Add new evaluator function",
      usage = "Enter a new function name and press enter to add it",
      order = -1,
      get = false,
      set = "AddEvaluatorFunction"
    }
  }

  self.configTable = {
    name = "GearEvaluator",
    desc = "Evaluates your gear based on certain rules",
    type = 'group',
    args = {
      evaluators = {
        name = "Evaluators",
        desc = "Evaluation functions",
        type = 'group',
        args = self.evaluatorConfigTable
      }
    }
  }

  self:UpdateEvaluatorConfigTable()
  self:SetConfigTable(self.configTable)
end

function GearEvaluator:CreateFrames()
  if (not self.currentlyModified) then
    self.currentlyModified = next(self.db.profile.evaluators)
  end
        
  local UpdateEvaluator = function()
    local text = GearEvaluatorModifyFrameEditBox:GetText()
    GearEvaluator.db.profile.evaluators[GearEvaluator.currentlyModified] = text
    GearEvaluator.evaluators[GearEvaluator.currentlyModified] = GearEvaluator:CompileEvaluator(text, GearEvaluator.currentlyModified)
    GearEvaluator:UpdateScore(GearEvaluator.currentlyModified)
  end
    
  if (not self.modifyFrame) then
    self.modifyFrame = CreateFrame("Frame", "GearEvaluatorModifyFrame", UIParent)
    self.modifyFrame:SetBackdrop({ bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", 
                                   edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", 
                                   tile = true, tileSize = 16, edgeSize = 16,
                                   insets = { left = 5, right = 6, top = 5, bottom = 6 }})
    self.modifyFrame:SetWidth(400)
    self.modifyFrame:SetHeight(300)
    self.modifyFrame:SetPoint("CENTER")
    self.modifyFrame:Hide()
    self.modifyFrame:SetScript("OnHide", UpdateEvaluator)
  end
    
  if (not self.modifyFrameCloseButton) then
    self.modifyFrameCloseButton = CreateFrame("Button", "GearEvaluatorModifyFrameCloseButton", self.modifyFrame, "UIPanelCloseButton")
    self.modifyFrameCloseButton:SetPoint("TOPRIGHT")
  end
  
  if (not self.modifyFrameEditBackground) then
    self.modifyFrameEditBackground = CreateFrame("Frame", "GearEvaluatorModifyFrameEditBackground", self.modifyFrame)
    self.modifyFrameEditBackground:SetBackdrop({ bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
                                                 edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
                                                 tile = true, tileSize = 16, edgeSize = 16,
                                                 insets = { left = 5, right = 6, top = 5, bottom = 6 }})
    self.modifyFrameEditBackground:SetPoint("LEFT", 5, 0)
    self.modifyFrameEditBackground:SetPoint("RIGHT", -5, 0)
    self.modifyFrameEditBackground:SetPoint("TOP", 0, -33)
    self.modifyFrameEditBackground:SetPoint("BOTTOM", 0, 5)    
		self.modifyFrameEditBackground:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b)
		self.modifyFrameEditBackground:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b)
  end
  
  if (not self.modifyFrameEditScroll) then
    self.modifyFrameEditScroll = CreateFrame("ScrollFrame", "GearEvaluatorModifyFrameEditScroll", self.modifyFrameEditBackground, "UIPanelScrollFrameTemplate")
    self.modifyFrameEditScroll:SetPoint("TOP", 0, -6)
    self.modifyFrameEditScroll:SetPoint("BOTTOM", 0, 6)
    self.modifyFrameEditScroll:SetPoint("LEFT", 6, 0)
    self.modifyFrameEditScroll:SetPoint("RIGHT", -28, 0)
  end
  
  if (not self.modifyFrameEditBox) then
    self.modifyFrameEditBox = CreateFrame("EditBox", "GearEvaluatorModifyFrameEditBox", self.modifyFrameEditScroll)
    self.modifyFrameEditScroll:SetScrollChild(self.modifyFrameEditBox)
    self.modifyFrameEditBox:SetWidth(self.modifyFrameEditScroll:GetWidth())
    self.modifyFrameEditBox:SetMultiLine(true)
    self.modifyFrameEditBox:SetText(self.db.profile.evaluators[self.currentlyModified])
    self.modifyFrameEditBox:SetFontObject(GameFontHighlightSmall)
    self.modifyFrameEditScroll:UpdateScrollChildRect()
    
    self.modifyFrameEditBox:SetScript("OnTextChanged", function()
      this:GetParent():UpdateScrollChildRect()
    end)
    
    self.modifyFrameEditBox:SetScript("OnCursorChanged", function(this, x, y, w, h)
      ScrollingEdit_OnCursorChanged(x, y, w, h)
      ScrollingEdit_OnUpdate()
    end)
  end
  
  if (not self.modifyFrameDropDownLabel) then
    self.modifyFrameDropDownLabel = self.modifyFrame:CreateFontString("GearEvaluatorModifyFrameDropDownLabel")
    self.modifyFrameDropDownLabel:SetFontObject(GameFontHighlightSmall)
    self.modifyFrameDropDownLabel:SetPoint("TOPLEFT", 13, -15)
    self.modifyFrameDropDownLabel:SetText("Modify evaluator:")
  end
  
  if (not self.modifyFrameDropDownMenu) then
    self.modifyFrameDropDownMenu = CreateFrame("Frame", "GearEvaluatorModifyFrameDropDownMenu", self.modifyFrame, "UIDropDownMenuTemplate")
    self.modifyFrameDropDownMenu:SetPoint("TOPLEFT", 90, -6)
    
    local SelectEvaluator = function()
      UpdateEvaluator()
      
      UIDropDownMenu_SetSelectedValue(this.owner, this.value)
      GearEvaluator.currentlyModified = this.value
      GearEvaluatorModifyFrameEditBox:SetText(GearEvaluator.db.profile.evaluators[this.value])
    end
    
    UIDropDownMenu_Initialize(self.modifyFrameDropDownMenu, function()
      local info = UIDropDownMenu_CreateInfo()

      for k, v in pairs(GearEvaluator.db.profile.evaluators) do
        info.value = k
        info.text = k
        info.owner = GearEvaluatorModifyFrameDropDownMenu
        info.func = SelectEvaluator
        info.checked = (GearEvaluator.currentlyModified == k)
        UIDropDownMenu_AddButton(info)
      end
    end)
    
    UIDropDownMenu_SetSelectedValue(self.modifyFrameDropDownMenu, self.currentlyModified)
  end
end

function GearEvaluator:UpdateEvaluatorConfigTable()
  -- Delete non-existing entries
  for k, v in pairs(self.evaluatorConfigTable) do
    if (not self.db.profile.evaluators[k] and k ~= "add") then
      self.evaluatorConfigTable[k] = nil
    end
  end

  -- Add new entries
  for k, v in pairs(self.db.profile.evaluators) do
    self.evaluatorConfigTable[k] = {
      type = 'group',
      name = k,
      desc = "Options for "..k,
      groupType = 'normal',
      args = {
        delete = {
          type = 'execute',
          name = "Delete",
          desc = "Click to delete the "..k.." evaluator",
          func = "DeleteEvaluatorFunction",
          passValue = k,
          confirmText = "Yes, really delete!"
        }
      }
    }
  end
end

function GearEvaluator:OnEnable()
  self:AddEventListener("UNIT_INVENTORY_CHANGED")
  self:AddEventListener("LibRockEvent-1.0", "FullyInitialized")

  self:CompileAllEvaluators()
end

function GearEvaluator:FullyInitialized()
  self:UpdateScores()
  self:RemoveEventListener("LibRockEvent-1.0", "FullyInitialized")
end

function GearEvaluator:OnDisable()
  self:RemoveEventListener("UNIT_INVENTORY_CHANGED")
  self:RemoveEventListener("LibRockEvent-1.0", "FullyInitialized")
end

function GearEvaluator:AddEvaluatorFunction(name)
  if (not self.db.profile.evaluators[name]) then
    self:Print("Adding evaluator "..name.."!")

    self.db.profile.evaluators[name] = "return 0"  
    self.evaluators[name] = self:CompileEvaluator(self.db.profile.evaluators[name])
    self:UpdateScore(name)
  
    self:UpdateEvaluatorConfigTable()
  else
    self:Print("Evaluator "..name.." is already defined!")
  end
end

function GearEvaluator:DeleteEvaluatorFunction(name)
  self.db.profile.evaluators[name] = nil
  self.evaluators[name] = nil
  self.scores[name] = nil

  self:UpdateEvaluatorConfigTable()
end

function GearEvaluator:CompileAllEvaluators()
  for k, v in pairs(self.db.profile.evaluators) do
    self.evaluators[k] = self:CompileEvaluator(v, k)
  end
end

function GearEvaluator:CompileEvaluator(text, name)
  local chunk, error = loadstring(text, name)

  if (not chunk) then
    self:Print(error)
    self.errors[name] = error
  end

  return chunk
end

function GearEvaluator:ShowModifyFrame()
 if (self.modifyFrame:IsShown()) then
    self.modifyFrame:Hide()
  else
    self.modifyFrame:Show()
  end
end

function GearEvaluator:OnUpdateFuBarTooltip()
  GameTooltip:AddLine("Current Scores:")

  for k in pairs(self.db.profile.evaluators) do
    GameTooltip:AddDoubleLine(k, self.scores[k] or self.errors[k])
  end

  GameTooltip:AddLine("Left-click to modify evaluators")
  GameTooltip:AddLine("Right-click to edit preferences")
end

function GearEvaluator:OnFuBarClick(button)
  self:ShowModifyFrame()
end

function GearEvaluator:GetEvaluatorEnvironment(gear)
  local evalenv = {}
  
  for name, module in self:IterateModulesWithMethod(true, "GetAttributes") do
    evalenv[name] = module:GetAttributes(gear)
  end
  
  return evalenv
end

function GearEvaluator:UpdateScore(name)
  if (not self.evaluators[name]) then
    self.scores[name] = nil
  else
    local evalenv = self:GetEvaluatorEnvironment(self:GetCurrentEquipment())
    local func = self.evaluators[name]

    setfenv(func, evalenv)
    
    local success, score = pcall(func)
    
    if (not success) then
      self:Print(score)
      self.scores[name] = nil
      self.errors[name] = score
    else
      self.scores[name] = score
    end
  end  
end

function GearEvaluator:UpdateScores()
  for k, v in pairs(self.evaluators) do
    self:UpdateScore(k)
  end
end

function GearEvaluator:GetCurrentEquipment()
  local gear = {}
  
  for k, s in pairs(slotnames) do
    gear[s] = GetInventoryItemLink("player", s)
  end
  
  return gear
end

-- Event handlers

function GearEvaluator:UNIT_INVENTORY_CHANGED(namespace, event, unit)
  if (unit == "player") then
    self:UpdateScores()
  end
end
